/*
 *  UltimateIRCd - an Internet Relay Chat Daemon, src/s_socks.c
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *  Refer to the documentation within doc/authors/ for full credits and copyrights.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: s_socks.c,v 1.24 2003/01/14 15:31:42 shadowmaster Exp $
 *
 */

/*
** Based on s_auth.c by Darren Reed (C) 1992,
** s_socks.c by Lucas Madar (C) 1998 and s_socks.c by
** ChatNet Coding Team.
*/


/* All these includes are copied from auth.c because we are
   just doing something very similar */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "version.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if defined(__hpux) || defined(DEBUGMODE)
#include "inet.h"
#endif
#include <fcntl.h>
#include "sock.h"		/*
				 * If FD_ZERO isn't define up to this point,
				 */
/*
 * define it (BSD4.2 needs this)
 */
#include "h.h"
#include "fdlist.h"

#ifdef SOCKSPORT


//static unsigned char socksid[12];
#define HICHAR(s)     (((unsigned short) s) >> 8)
#define LOCHAR(s)     (((unsigned short) s) & 0xFF)
#define LISTEN_SIZE 5

/*
 * start_proxy
 *
 * attempts to connect to a socks4 proxy server on port 1080 of the host.
 * if the connect fails, the user passes the proxy check
 * otherwise, the connection is checked to see if it is a "secure"
 * socks4+ server
 */
void
start_proxy (aClient * client_p)
{
  struct SOCKADDR_IN sock;
  struct SOCKADDR_IN localaddr;
  int locallen;

  Debug ((DEBUG_NOTICE, "start_proxy(%x) fd %d status %d",
	  client_p, client_p->fd, client_p->status));

  if ((client_p->proxyfd = socket (AFINET, SOCK_STREAM, 0)) == -1)
    {
#ifdef	USE_SYSLOG
      syslog (LOG_ERR, "Unable to create proxy check socket for %s:%m",
	      get_client_name (client_p, TRUE));
#endif
      return;
    }
  if (client_p->proxyfd >= MAXCONNECTIONS)
    {
      sendto_ops ("Can't allocate fd for proxy check on %s",
		  get_client_name (client_p,
				   (IsServer (client_p) ? HIDEME : TRUE)));

      close (client_p->proxyfd);
      client_p->proxyfd = -1;
      return;
    }
#ifdef SHOW_HEADERS
#ifdef USE_SSL
  if (!IsSSL (client_p))
    {
#endif
      sendto_one (client_p, REPORT_DO_PROXY);
#ifdef USE_SSL
    }
#endif
#endif

  set_non_blocking (client_p->proxyfd, client_p);
  /*
   * get the local address of the client and bind to that to make the
   * proxy request.  This used to be done only for ifdef VIRTTUAL_HOST,
   * but needs to be done for all clients since the ident request must
   * originate from that same address-- and machines with multiple IP
   * addresses are common now
   */
  locallen = sizeof (struct SOCKADDR_IN);

  memset (&localaddr, '\0', locallen);
  getsockname (client_p->fd, (struct SOCKADDR *) &localaddr, &locallen);
  localaddr.SIN_PORT = htons (0);
  if (bind (client_p->proxyfd, (struct SOCKADDR *) &localaddr,
	    sizeof (localaddr)) == -1)
    {
      report_error ("binding proxy stream socket %s:%s", client_p);
      close (client_p->proxyfd);
      client_p->proxyfd = -1;
      return;
    }

  memcpy ((char *) &sock.SIN_ADDR, (char *) &client_p->ip,
	  sizeof (struct IN_ADDR));

  sock.SIN_PORT = htons (1080);
  sock.SIN_FAMILY = AFINET;
  /*
   * (void)alarm((unsigned)4);
   */
  if (connect (client_p->proxyfd, (struct SOCKADDR *) &sock,
	       sizeof (sock)) == -1 && errno != EINPROGRESS)
    {
      /*
       * No error report from this...
       */
      /*
       * (void)alarm((unsigned)0);
       */
      close (client_p->proxyfd);
      client_p->proxyfd = -1;
#ifdef SHOW_HEADERS
#ifdef USE_SSL
      if (!IsSSL (client_p))
	{
#endif
	  sendto_one (client_p, REPORT_NO_PROXY);
#ifdef USE_SSL
	}
#endif
#endif
      return;
    }
  /*
   * (void)alarm((unsigned)0);
   */
  client_p->flags2 |= (FLAGS2_WRPROXY | FLAGS2_PROXY);
  if (client_p->proxyfd > highest_fd)
    highest_fd = client_p->proxyfd;
  return;
}


/*
 * send_proxyquery
 *
 * send the proxy server a query to see if it's open.
 */
void
send_proxyquery (aClient * client_p)
{
  struct SOCKADDR_IN us, them;
  char proxybuf[12];
  int ulen, tlen;
  unsigned long theip;

  Debug ((DEBUG_NOTICE, "write_proxyports(%x) fd %d proxyfd %d stat %d",
	  client_p, client_p->fd, client_p->proxyfd, client_p->status));
  tlen = ulen = sizeof (us);

  if (getsockname (client_p->fd, (struct SOCKADDR *) &us, &ulen) ||
      getpeername (client_p->fd, (struct SOCKADDR *) &them, &tlen))
    {
#ifdef	USE_SYSLOG
      syslog (LOG_DEBUG, "proxy get{sock,peer}name error for %s:%m",
	      get_client_name (client_p, TRUE));
#endif
      close (client_p->proxyfd);
      if (client_p->proxyfd == highest_fd)
	while (!local[highest_fd])
	  highest_fd--;
      client_p->proxyfd = -1;
      client_p->flags2 &= ~(FLAGS2_PROXY | FLAGS2_WRPROXY);
      return;
    }
/*** FIXME ***/
/* Will this work properly under IPv6? */
  theip = htonl (us.SIN_ADDR.S_ADDR);

  proxybuf[0] = 4;		// socks 4
  proxybuf[1] = 1;		// connect request
  proxybuf[2] = ((((unsigned short) ntohs (us.SIN_PORT)) >> 8) & 0xFF);	// hiword of port
  proxybuf[3] = (((unsigned short) ntohs (us.SIN_PORT)) & 0xFF);	// loword of port
  proxybuf[4] = (theip >> 24) & 0xFF;
  proxybuf[5] = (theip >> 16) & 0xFF;
  proxybuf[6] = (theip >> 8) & 0xFF;
  proxybuf[7] = (theip & 0xFF);
  proxybuf[8] = 0;

  Debug ((DEBUG_SEND, "sending [%s] to proxy port %s.1080",
	  proxybuf, inet_ntop (AFINET, (char *) &them.SIN_ADDR, mydummy,
			       sizeof (mydummy))));

  if (write (client_p->proxyfd, proxybuf, strlen (proxybuf)) !=
      strlen (proxybuf))
    {
      close (client_p->proxyfd);
      if (client_p->proxyfd == highest_fd)
	while (!local[highest_fd])
	  highest_fd--;
      client_p->proxyfd = -1;
      client_p->flags2 &= ~FLAGS2_PROXY;
      send (client_p->fd, REPORT_NO_PROXY);
      return;
    }

  client_p->flags2 &= ~FLAGS2_WRPROXY;

  return;
}

/*
 * read_proxy
 *
 * process the proxy reply.
 */

void
read_proxy (aClient * client_p)
{
  unsigned char proxybuf[12];
  int len;

  len = RECV (client_p->proxyfd, proxybuf, 9);

  Debug ((DEBUG_NOTICE, "read_proxy: %s", proxybuf));

  close (client_p->proxyfd);
  if (client_p->proxyfd == highest_fd)
    while (!local[highest_fd])
      highest_fd--;
  client_p->proxyfd = -1;
  ClearProxy (client_p);

  if (len < 4)
    {
#ifdef USE_SSL
      if (!IsSSL (client_p))
#endif
	SEND (client_p->fd, REPORT_NO_PROXY);
      return;
    }
  if (proxybuf[1] == 90)
    {
      client_p->flags2 |= FLAGS2_GOTPROXY;
      return;
    }
#ifdef SHOW_HEADERS
#ifdef USE_SSL
  if (!IsSSL (client_p))
    {
#endif
      sendto_one (client_p, REPORT_GOOD_PROXY);
#ifdef USE_SSL
    }
#endif
#endif
  return;
}

#endif /* SOCKSPORT */
